/* knapuser.c includes file bbskel.c
 *
 * This fork95 program solves the 0-1 Knapsack problem using the
 * Branch and Bound skeleton prototype specified in file bbskel.h
 * For information on Branch and Bound and the Knapsack Problem
 * you can use the book of Horowitz and Sahni,
 * "Fundamentals of Computers Algorithms" (Pitman)
 * For more detailed information about the Knapsack problem, see
 * "Knapsack Problems. Algorithms and Computer Implementations." by
 * Silvano Martello and Paolo Toth. John Wiley and Sons.
 * This approach focuses in the aspects of general
 * parallel branch and bound and does not in the concrete
 * features of knapsack Problems.
 * The  user has to specify the same items than in a sequential 
 * branch and bound algorithm. That is:
 * 1) the type defining a problem. Has to be called "problemt"
 * 2) the type "number" returned for the function to optimize
 * 3) the type "index" of the number of problems in the heap 
 *   (usually an "int" will be sufficient)
 * 4) the function setProblem:
 *              void setProblem(problemt *Prob, ...) 
 *    this functions return in Prob the problem built from the data
 *    constituting the subsequent parameters.
 * 5) the function numChildren: 
 *              index numChildren(problemt Prob) 
 *    returning the number of children of problem Prob.
 * 6) the function setChild:
 *      void setChild(problemt *child, problemt father,index order) 
 *    it returns in "child" the "order"-th child problem 
 *    among the children of problem "father".
 * 7) the function
 *      void lowerUpper(problemt Prob, number *L, number *U) 
 *    it returns in L the lower bound and in U the upper 
 *    bound of problem "Prob".
 * 8) The variable "bestSol" of type "number". It will contain
 *    the solution to the problem.
 * the program is only a class example and does not intend 
 * to be used with other purposes.
 * The skeletton provides the functions
 *
 *			insertProblem(Prob)
 * 	  	branchAndBound()
 * the first one inserts a problem in the heap, the second 
 * solves the problem using the branch and bound technique.
 * Once the functions defining the b&b structure has been provided
 * the user has to insert the problem in the heap and call to 
 *
 * branch and bound function (see main). 
 * TODO LIST 
 * (THIS IS A FIRST ATTEMPT TO DEFINE A LIST OF TASKS FOR MARCELA AND
 * FABIANA. I RATHER PREFER BOTH OF YOU TO START TO WORK TOGETHER
 * IN THE SAME SUBJECT, WE WILL DIFFERENTIATE IN THE NEAR FUTURE.)
 *
 * 
 * 1) #define KINDOFPROBLEM 
 *    it could take as values MAX or MIN 
 *    to indicate if the problem is a maximization or minimization 
 *    problem.
 * 2) prefix bb to all the variables and functions in the skeletton
 *    to avoid crashes and conflicts with user variables.
 * 3) #define SEARCHPOLICY. Values will be DEPTH, BEST, BREADTH.
 *    implement correspondent branch and bound routines.
 * 4) to allow the user different names of the b&b defining functions,
 *    change user defined functions to macros. The user defines 
 *    the names of the different functions and variables so that 
 *    several branch and bound s can be called inside a program 
 *    using #undef
 *    #define LOWERUPPER lowerUpper
 *    #define SETCHILD setChild
 *    etc.
 * 4) #include undefines.h 
 *    this file will undefine all the elements related with 
 *    the branch and bound skeleton using the directive #undef.
 *
 * 5) Write a TSP branch and bound using the skeleton
 * 6) Write a program that solves first the knapsack and after 
 *    the TSP.
 * 7) Do this approach using objects
 * 8) Use MPI as target. Study BSP as target. Study HPF as target.
 * 9) Built an analogous skeletton for divide and conquer d&c 
 * 10) using the automaton approach try to built an skeletton 
 *     for Dynamic Programming (DP).
 * 11) Consider the Multidimensional Knapsack Problem (MKP). To solve 
 *     it is required to use d&c (for the sorting of objects)
 *     and several iterative nested b&b. To find the Upper Bound, 
 *     a linear combination of the restrictions  reduces the MKP to
 *     the 1-KNP.
 *     Use the skeletons to find the solution.
 * 12) Implement the whole thing using objects and OOP.
 * 13) Built a graphical interface for the skelettons.
 * Author: Casiano Rodriguez Leon. Email: casiano@ull.es
 * March, 1999
*/

#include <fork.h>
#include <io.h>
#include <stdlib.h>
#include <assert.h>

typedef int index;
typedef int number;
typedef int boolean;

typedef struct {
	number k, C, P;
} problemt;

#define MAXNUMBEROFOBJECTS 64 
#define STRIDE 1
#define  INFINITY 8192
/* INFINITY has to be set to a value greater 
   than the capacity M of the Knapsack */

#define ns "%d\n"

/* knapsack support variables */
sh number N, M;
sh number w[MAXNUMBEROFOBJECTS], p[MAXNUMBEROFOBJECTS];
sh number bestSol = -INFINITY; 

sync void readKnap(void) {
#define STRIDE 1
	FILE *in;
  index i;

	farm {
		if ((in = fopen("knap.dat","r+")) == NULL) {
		 fprintf(stderr, "Cannot open input file knap.dat.\n");
		 fprintf(stderr,"Input file format is:\nNumber of objects\nCapacity\nweight 0\nprofit 0\n...\n");
		 exit(1);
		}
		fscanf(in,ns,&N);
		fscanf(in,ns,&M);
		printf("N = "); printf(ns,N);
		printf("M = "); printf(ns,M);
		assert(INFINITY > M);
		for(i=0;i<N;i++) {
			fscanf(in,ns,&w[i]);
			fscanf(in,ns,&p[i]);
			if (i%STRIDE == 0) {
				printf("w = "); printf(ns,w[i]);
				printf("p = "); printf(ns,p[i]);
		 }
		}
		w[N] = INFINITY; p[N] = 0; /* sentinel object */
		fclose(in);
	} /* farm */
}

void lowerUpper(problemt Prob, number *L, number *U) {
  /* naive approach */
  number i, weig = 0, prof = Prob.P;

  if (Prob.C < 0) {
	 *L = -INFINITY; *U = -INFINITY;
  }
  else {
	 for(i=Prob.k; weig <= Prob.C; i++) {
		weig += w[i]; prof += p[i];
	 }
	 /* invariant:  weig > prob.C */
	 i--;
	 weig -= w[i];
	 prof -= p[i];
	 *L = prof;  /* see Martello and Toth pag 16. Th 2.1 */
	 *U = prof+(p[i]*(Prob.C-weig))/w[i];
  }
}

void setProblem(problemt *Prob, number k, number C, number P) {

  Prob->k = k;
  Prob->C = C;
  Prob->P = P;
}

index numChildren(problemt Prob) {
  if (Prob.k < N) return 2; else return 0;
}

void setChild(problemt *child, problemt father,index order) {

  setProblem(child,
				 father.k+1,
				 father.C-(1-order)*w[father.k],
				 father.P+(1-order)*p[father.k]
				);
}

/********* skeleton must be included after the  definition of ****/
/********* the instance functions. *******************************/

#include "bbskel.h"

int main(void) {
  problemt Prob;
	int i;

	start {
		if ($ ==0) {
			readKnap();
			farm setProblem(&Prob, 0, M, 0);
			insertProblem(Prob);
		}

  	branchAndBound();

		NUMPROCESSORS = 0;
    $ = mpadd(&NUMPROCESSORS,1);
		for(i=0; i < NUMPROCESSORS; i++) 
			if ($ == i) seq { printf("\nsol = "); printf(ns, bestSol);}
  } /* end start */
}

